# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import typing_extensions
from typing import Dict, List, Union, Iterable, Optional
from datetime import datetime
from typing_extensions import Literal, overload

import httpx

from ..types import (
    Currency,
    SubscriptionStatus,
    subscription_list_params,
    subscription_charge_params,
    subscription_create_params,
    subscription_update_params,
    subscription_change_plan_params,
    subscription_preview_change_plan_params,
    subscription_update_payment_method_params,
    subscription_retrieve_usage_history_params,
)
from .._types import Body, Omit, Query, Headers, NoneType, NotGiven, omit, not_given
from .._utils import required_args, maybe_transform, async_maybe_transform
from .._compat import cached_property
from .._resource import SyncAPIResource, AsyncAPIResource
from .._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ..pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination
from .._base_client import AsyncPaginator, make_request_options
from ..types.currency import Currency
from ..types.subscription import Subscription
from ..types.attach_addon_param import AttachAddonParam
from ..types.subscription_status import SubscriptionStatus
from ..types.payment_method_types import PaymentMethodTypes
from ..types.billing_address_param import BillingAddressParam
from ..types.customer_request_param import CustomerRequestParam
from ..types.subscription_list_response import SubscriptionListResponse
from ..types.on_demand_subscription_param import OnDemandSubscriptionParam
from ..types.subscription_charge_response import SubscriptionChargeResponse
from ..types.subscription_create_response import SubscriptionCreateResponse
from ..types.subscription_preview_change_plan_response import SubscriptionPreviewChangePlanResponse
from ..types.subscription_retrieve_credit_usage_response import SubscriptionRetrieveCreditUsageResponse
from ..types.subscription_update_payment_method_response import SubscriptionUpdatePaymentMethodResponse
from ..types.subscription_retrieve_usage_history_response import SubscriptionRetrieveUsageHistoryResponse

__all__ = ["SubscriptionsResource", "AsyncSubscriptionsResource"]


class SubscriptionsResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> SubscriptionsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return SubscriptionsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> SubscriptionsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return SubscriptionsResourceWithStreamingResponse(self)

    @typing_extensions.deprecated("deprecated")
    def create(
        self,
        *,
        billing: BillingAddressParam,
        customer: CustomerRequestParam,
        product_id: str,
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        allowed_payment_method_types: Optional[List[PaymentMethodTypes]] | Omit = omit,
        billing_currency: Optional[Currency] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        force_3ds: Optional[bool] | Omit = omit,
        metadata: Dict[str, str] | Omit = omit,
        on_demand: Optional[OnDemandSubscriptionParam] | Omit = omit,
        one_time_product_cart: Optional[Iterable[subscription_create_params.OneTimeProductCart]] | Omit = omit,
        payment_link: Optional[bool] | Omit = omit,
        payment_method_id: Optional[str] | Omit = omit,
        redirect_immediately: bool | Omit = omit,
        return_url: Optional[str] | Omit = omit,
        short_link: Optional[bool] | Omit = omit,
        show_saved_payment_methods: bool | Omit = omit,
        tax_id: Optional[str] | Omit = omit,
        trial_period_days: Optional[int] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionCreateResponse:
        """
        Args:
          billing: Billing address information for the subscription

          customer: Customer details for the subscription

          product_id: Unique identifier of the product to subscribe to

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Attach addons to this subscription

          allowed_payment_method_types: List of payment methods allowed during checkout.

              Customers will **never** see payment methods that are **not** in this list.
              However, adding a method here **does not guarantee** customers will see it.
              Availability still depends on other factors (e.g., customer location, merchant
              settings).

          billing_currency: Fix the currency in which the end customer is billed. If Dodo Payments cannot
              support that currency for this transaction, it will not proceed

          discount_code: Discount Code to apply to the subscription

          force_3ds: Override merchant default 3DS behaviour for this subscription

          metadata: Additional metadata for the subscription Defaults to empty if not specified

          one_time_product_cart: List of one time products that will be bundled with the first payment for this
              subscription

          payment_link: If true, generates a payment link. Defaults to false if not specified.

          payment_method_id: Optional payment method ID to use for this subscription. If provided,
              customer_id must also be provided (via AttachExistingCustomer). The payment
              method will be validated for eligibility with the subscription's currency.

          redirect_immediately: If true, redirects the customer immediately after payment completion False by
              default

          return_url: Optional URL to redirect after successful subscription creation

          short_link: If true, returns a shortened payment link. Defaults to false if not specified.

          show_saved_payment_methods: Display saved payment methods of a returning customer False by default

          tax_id: Tax ID in case the payment is B2B. If tax id validation fails the payment
              creation will fail

          trial_period_days: Optional trial period in days If specified, this value overrides the trial
              period set in the product's price Must be between 0 and 10000 days

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._post(
            "/subscriptions",
            body=maybe_transform(
                {
                    "billing": billing,
                    "customer": customer,
                    "product_id": product_id,
                    "quantity": quantity,
                    "addons": addons,
                    "allowed_payment_method_types": allowed_payment_method_types,
                    "billing_currency": billing_currency,
                    "discount_code": discount_code,
                    "force_3ds": force_3ds,
                    "metadata": metadata,
                    "on_demand": on_demand,
                    "one_time_product_cart": one_time_product_cart,
                    "payment_link": payment_link,
                    "payment_method_id": payment_method_id,
                    "redirect_immediately": redirect_immediately,
                    "return_url": return_url,
                    "short_link": short_link,
                    "show_saved_payment_methods": show_saved_payment_methods,
                    "tax_id": tax_id,
                    "trial_period_days": trial_period_days,
                },
                subscription_create_params.SubscriptionCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionCreateResponse,
        )

    def retrieve(
        self,
        subscription_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Subscription:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._get(
            f"/subscriptions/{subscription_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Subscription,
        )

    def update(
        self,
        subscription_id: str,
        *,
        billing: Optional[BillingAddressParam] | Omit = omit,
        cancel_at_next_billing_date: Optional[bool] | Omit = omit,
        credit_entitlement_cart: Optional[Iterable[subscription_update_params.CreditEntitlementCart]] | Omit = omit,
        customer_name: Optional[str] | Omit = omit,
        disable_on_demand: Optional[subscription_update_params.DisableOnDemand] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        next_billing_date: Union[str, datetime, None] | Omit = omit,
        status: Optional[SubscriptionStatus] | Omit = omit,
        tax_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Subscription:
        """
        Args:
          cancel_at_next_billing_date: When set, the subscription will remain active until the end of billing period

          credit_entitlement_cart: Update credit entitlement cart settings

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._patch(
            f"/subscriptions/{subscription_id}",
            body=maybe_transform(
                {
                    "billing": billing,
                    "cancel_at_next_billing_date": cancel_at_next_billing_date,
                    "credit_entitlement_cart": credit_entitlement_cart,
                    "customer_name": customer_name,
                    "disable_on_demand": disable_on_demand,
                    "metadata": metadata,
                    "next_billing_date": next_billing_date,
                    "status": status,
                    "tax_id": tax_id,
                },
                subscription_update_params.SubscriptionUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Subscription,
        )

    def list(
        self,
        *,
        brand_id: str | Omit = omit,
        created_at_gte: Union[str, datetime] | Omit = omit,
        created_at_lte: Union[str, datetime] | Omit = omit,
        customer_id: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        product_id: str | Omit = omit,
        status: Literal["pending", "active", "on_hold", "cancelled", "failed", "expired"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[SubscriptionListResponse]:
        """
        Args:
          brand_id: filter by Brand id

          created_at_gte: Get events after this created time

          created_at_lte: Get events created before this time

          customer_id: Filter by customer id

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          product_id: Filter by product id

          status: Filter by status

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/subscriptions",
            page=SyncDefaultPageNumberPagination[SubscriptionListResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "brand_id": brand_id,
                        "created_at_gte": created_at_gte,
                        "created_at_lte": created_at_lte,
                        "customer_id": customer_id,
                        "page_number": page_number,
                        "page_size": page_size,
                        "product_id": product_id,
                        "status": status,
                    },
                    subscription_list_params.SubscriptionListParams,
                ),
            ),
            model=SubscriptionListResponse,
        )

    def change_plan(
        self,
        subscription_id: str,
        *,
        product_id: str,
        proration_billing_mode: Literal["prorated_immediately", "full_immediately", "difference_immediately"],
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        on_payment_failure: Optional[Literal["prevent_change", "apply_change"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> None:
        """
        Args:
          product_id: Unique identifier of the product to subscribe to

          proration_billing_mode: Proration Billing Mode

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Addons for the new plan. Note : Leaving this empty would remove any existing
              addons

          discount_code: Optional discount code to apply to the new plan. If provided, validates and
              applies the discount to the plan change. If not provided and the subscription
              has an existing discount with `preserve_on_plan_change=true`, the existing
              discount will be preserved (if applicable to the new product).

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          on_payment_failure: Controls behavior when the plan change payment fails.

              - `prevent_change`: Keep subscription on current plan until payment succeeds
              - `apply_change` (default): Apply plan change immediately regardless of payment
                outcome

              If not specified, uses the business-level default setting.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        extra_headers = {"Accept": "*/*", **(extra_headers or {})}
        return self._post(
            f"/subscriptions/{subscription_id}/change-plan",
            body=maybe_transform(
                {
                    "product_id": product_id,
                    "proration_billing_mode": proration_billing_mode,
                    "quantity": quantity,
                    "addons": addons,
                    "discount_code": discount_code,
                    "metadata": metadata,
                    "on_payment_failure": on_payment_failure,
                },
                subscription_change_plan_params.SubscriptionChangePlanParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=NoneType,
        )

    def charge(
        self,
        subscription_id: str,
        *,
        product_price: int,
        adaptive_currency_fees_inclusive: Optional[bool] | Omit = omit,
        customer_balance_config: Optional[subscription_charge_params.CustomerBalanceConfig] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        product_currency: Optional[Currency] | Omit = omit,
        product_description: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionChargeResponse:
        """Args:
          product_price: The product price.

        Represented in the lowest denomination of the currency (e.g.,
              cents for USD). For example, to charge $1.00, pass `100`.

          adaptive_currency_fees_inclusive: Whether adaptive currency fees should be included in the product_price (true) or
              added on top (false). This field is ignored if adaptive pricing is not enabled
              for the business.

          customer_balance_config: Specify how customer balance is used for the payment

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          product_currency: Optional currency of the product price. If not specified, defaults to the
              currency of the product.

          product_description: Optional product description override for billing and line items. If not
              specified, the stored description of the product will be used.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._post(
            f"/subscriptions/{subscription_id}/charge",
            body=maybe_transform(
                {
                    "product_price": product_price,
                    "adaptive_currency_fees_inclusive": adaptive_currency_fees_inclusive,
                    "customer_balance_config": customer_balance_config,
                    "metadata": metadata,
                    "product_currency": product_currency,
                    "product_description": product_description,
                },
                subscription_charge_params.SubscriptionChargeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionChargeResponse,
        )

    def preview_change_plan(
        self,
        subscription_id: str,
        *,
        product_id: str,
        proration_billing_mode: Literal["prorated_immediately", "full_immediately", "difference_immediately"],
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        on_payment_failure: Optional[Literal["prevent_change", "apply_change"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionPreviewChangePlanResponse:
        """
        Args:
          product_id: Unique identifier of the product to subscribe to

          proration_billing_mode: Proration Billing Mode

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Addons for the new plan. Note : Leaving this empty would remove any existing
              addons

          discount_code: Optional discount code to apply to the new plan. If provided, validates and
              applies the discount to the plan change. If not provided and the subscription
              has an existing discount with `preserve_on_plan_change=true`, the existing
              discount will be preserved (if applicable to the new product).

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          on_payment_failure: Controls behavior when the plan change payment fails.

              - `prevent_change`: Keep subscription on current plan until payment succeeds
              - `apply_change` (default): Apply plan change immediately regardless of payment
                outcome

              If not specified, uses the business-level default setting.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._post(
            f"/subscriptions/{subscription_id}/change-plan/preview",
            body=maybe_transform(
                {
                    "product_id": product_id,
                    "proration_billing_mode": proration_billing_mode,
                    "quantity": quantity,
                    "addons": addons,
                    "discount_code": discount_code,
                    "metadata": metadata,
                    "on_payment_failure": on_payment_failure,
                },
                subscription_preview_change_plan_params.SubscriptionPreviewChangePlanParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionPreviewChangePlanResponse,
        )

    def retrieve_credit_usage(
        self,
        subscription_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionRetrieveCreditUsageResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._get(
            f"/subscriptions/{subscription_id}/credit-usage",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionRetrieveCreditUsageResponse,
        )

    def retrieve_usage_history(
        self,
        subscription_id: str,
        *,
        end_date: Union[str, datetime, None] | Omit = omit,
        meter_id: Optional[str] | Omit = omit,
        page_number: Optional[int] | Omit = omit,
        page_size: Optional[int] | Omit = omit,
        start_date: Union[str, datetime, None] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse]:
        """
        Get detailed usage history for a subscription that includes usage-based billing
        (metered components). This endpoint provides insights into customer usage
        patterns and billing calculations over time.

        ## What You'll Get:

        - **Billing periods**: Each item represents a billing cycle with start and end
          dates
        - **Meter usage**: Detailed breakdown of usage for each meter configured on the
          subscription
        - **Usage calculations**: Total units consumed, free threshold units, and
          chargeable units
        - **Historical tracking**: Complete audit trail of usage-based charges

        ## Use Cases:

        - **Customer support**: Investigate billing questions and usage discrepancies
        - **Usage analytics**: Analyze customer consumption patterns over time
        - **Billing transparency**: Provide customers with detailed usage breakdowns
        - **Revenue optimization**: Identify usage trends to optimize pricing strategies

        ## Filtering Options:

        - **Date range filtering**: Get usage history for specific time periods
        - **Meter-specific filtering**: Focus on usage for a particular meter
        - **Pagination**: Navigate through large usage histories efficiently

        ## Important Notes:

        - Only returns data for subscriptions with usage-based (metered) components
        - Usage history is organized by billing periods (subscription cycles)
        - Free threshold units are calculated and displayed separately from chargeable
          units
        - Historical data is preserved even if meter configurations change

        ## Example Query Patterns:

        - Get last 3 months:
          `?start_date=2024-01-01T00:00:00Z&end_date=2024-03-31T23:59:59Z`
        - Filter by meter: `?meter_id=mtr_api_requests`
        - Paginate results: `?page_size=20&page_number=1`
        - Recent usage: `?start_date=2024-03-01T00:00:00Z` (from March 1st to now)

        Args:
          end_date: Filter by end date (inclusive)

          meter_id: Filter by specific meter ID

          page_number: Page number (default: 0)

          page_size: Page size (default: 10, max: 100)

          start_date: Filter by start date (inclusive)

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._get_api_list(
            f"/subscriptions/{subscription_id}/usage-history",
            page=SyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "end_date": end_date,
                        "meter_id": meter_id,
                        "page_number": page_number,
                        "page_size": page_size,
                        "start_date": start_date,
                    },
                    subscription_retrieve_usage_history_params.SubscriptionRetrieveUsageHistoryParams,
                ),
            ),
            model=SubscriptionRetrieveUsageHistoryResponse,
        )

    @overload
    def update_payment_method(
        self,
        subscription_id: str,
        *,
        type: Literal["new"],
        return_url: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        ...

    @overload
    def update_payment_method(
        self,
        subscription_id: str,
        *,
        payment_method_id: str,
        type: Literal["existing"],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        ...

    @required_args(["type"], ["payment_method_id", "type"])
    def update_payment_method(
        self,
        subscription_id: str,
        *,
        type: Literal["new"] | Literal["existing"],
        return_url: Optional[str] | Omit = omit,
        payment_method_id: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._post(
            f"/subscriptions/{subscription_id}/update-payment-method",
            body=maybe_transform(
                {
                    "type": type,
                    "return_url": return_url,
                    "payment_method_id": payment_method_id,
                },
                subscription_update_payment_method_params.SubscriptionUpdatePaymentMethodParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionUpdatePaymentMethodResponse,
        )


class AsyncSubscriptionsResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncSubscriptionsResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncSubscriptionsResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncSubscriptionsResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncSubscriptionsResourceWithStreamingResponse(self)

    @typing_extensions.deprecated("deprecated")
    async def create(
        self,
        *,
        billing: BillingAddressParam,
        customer: CustomerRequestParam,
        product_id: str,
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        allowed_payment_method_types: Optional[List[PaymentMethodTypes]] | Omit = omit,
        billing_currency: Optional[Currency] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        force_3ds: Optional[bool] | Omit = omit,
        metadata: Dict[str, str] | Omit = omit,
        on_demand: Optional[OnDemandSubscriptionParam] | Omit = omit,
        one_time_product_cart: Optional[Iterable[subscription_create_params.OneTimeProductCart]] | Omit = omit,
        payment_link: Optional[bool] | Omit = omit,
        payment_method_id: Optional[str] | Omit = omit,
        redirect_immediately: bool | Omit = omit,
        return_url: Optional[str] | Omit = omit,
        short_link: Optional[bool] | Omit = omit,
        show_saved_payment_methods: bool | Omit = omit,
        tax_id: Optional[str] | Omit = omit,
        trial_period_days: Optional[int] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionCreateResponse:
        """
        Args:
          billing: Billing address information for the subscription

          customer: Customer details for the subscription

          product_id: Unique identifier of the product to subscribe to

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Attach addons to this subscription

          allowed_payment_method_types: List of payment methods allowed during checkout.

              Customers will **never** see payment methods that are **not** in this list.
              However, adding a method here **does not guarantee** customers will see it.
              Availability still depends on other factors (e.g., customer location, merchant
              settings).

          billing_currency: Fix the currency in which the end customer is billed. If Dodo Payments cannot
              support that currency for this transaction, it will not proceed

          discount_code: Discount Code to apply to the subscription

          force_3ds: Override merchant default 3DS behaviour for this subscription

          metadata: Additional metadata for the subscription Defaults to empty if not specified

          one_time_product_cart: List of one time products that will be bundled with the first payment for this
              subscription

          payment_link: If true, generates a payment link. Defaults to false if not specified.

          payment_method_id: Optional payment method ID to use for this subscription. If provided,
              customer_id must also be provided (via AttachExistingCustomer). The payment
              method will be validated for eligibility with the subscription's currency.

          redirect_immediately: If true, redirects the customer immediately after payment completion False by
              default

          return_url: Optional URL to redirect after successful subscription creation

          short_link: If true, returns a shortened payment link. Defaults to false if not specified.

          show_saved_payment_methods: Display saved payment methods of a returning customer False by default

          tax_id: Tax ID in case the payment is B2B. If tax id validation fails the payment
              creation will fail

          trial_period_days: Optional trial period in days If specified, this value overrides the trial
              period set in the product's price Must be between 0 and 10000 days

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return await self._post(
            "/subscriptions",
            body=await async_maybe_transform(
                {
                    "billing": billing,
                    "customer": customer,
                    "product_id": product_id,
                    "quantity": quantity,
                    "addons": addons,
                    "allowed_payment_method_types": allowed_payment_method_types,
                    "billing_currency": billing_currency,
                    "discount_code": discount_code,
                    "force_3ds": force_3ds,
                    "metadata": metadata,
                    "on_demand": on_demand,
                    "one_time_product_cart": one_time_product_cart,
                    "payment_link": payment_link,
                    "payment_method_id": payment_method_id,
                    "redirect_immediately": redirect_immediately,
                    "return_url": return_url,
                    "short_link": short_link,
                    "show_saved_payment_methods": show_saved_payment_methods,
                    "tax_id": tax_id,
                    "trial_period_days": trial_period_days,
                },
                subscription_create_params.SubscriptionCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionCreateResponse,
        )

    async def retrieve(
        self,
        subscription_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Subscription:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._get(
            f"/subscriptions/{subscription_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Subscription,
        )

    async def update(
        self,
        subscription_id: str,
        *,
        billing: Optional[BillingAddressParam] | Omit = omit,
        cancel_at_next_billing_date: Optional[bool] | Omit = omit,
        credit_entitlement_cart: Optional[Iterable[subscription_update_params.CreditEntitlementCart]] | Omit = omit,
        customer_name: Optional[str] | Omit = omit,
        disable_on_demand: Optional[subscription_update_params.DisableOnDemand] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        next_billing_date: Union[str, datetime, None] | Omit = omit,
        status: Optional[SubscriptionStatus] | Omit = omit,
        tax_id: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> Subscription:
        """
        Args:
          cancel_at_next_billing_date: When set, the subscription will remain active until the end of billing period

          credit_entitlement_cart: Update credit entitlement cart settings

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._patch(
            f"/subscriptions/{subscription_id}",
            body=await async_maybe_transform(
                {
                    "billing": billing,
                    "cancel_at_next_billing_date": cancel_at_next_billing_date,
                    "credit_entitlement_cart": credit_entitlement_cart,
                    "customer_name": customer_name,
                    "disable_on_demand": disable_on_demand,
                    "metadata": metadata,
                    "next_billing_date": next_billing_date,
                    "status": status,
                    "tax_id": tax_id,
                },
                subscription_update_params.SubscriptionUpdateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=Subscription,
        )

    def list(
        self,
        *,
        brand_id: str | Omit = omit,
        created_at_gte: Union[str, datetime] | Omit = omit,
        created_at_lte: Union[str, datetime] | Omit = omit,
        customer_id: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        product_id: str | Omit = omit,
        status: Literal["pending", "active", "on_hold", "cancelled", "failed", "expired"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[SubscriptionListResponse, AsyncDefaultPageNumberPagination[SubscriptionListResponse]]:
        """
        Args:
          brand_id: filter by Brand id

          created_at_gte: Get events after this created time

          created_at_lte: Get events created before this time

          customer_id: Filter by customer id

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          product_id: Filter by product id

          status: Filter by status

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/subscriptions",
            page=AsyncDefaultPageNumberPagination[SubscriptionListResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "brand_id": brand_id,
                        "created_at_gte": created_at_gte,
                        "created_at_lte": created_at_lte,
                        "customer_id": customer_id,
                        "page_number": page_number,
                        "page_size": page_size,
                        "product_id": product_id,
                        "status": status,
                    },
                    subscription_list_params.SubscriptionListParams,
                ),
            ),
            model=SubscriptionListResponse,
        )

    async def change_plan(
        self,
        subscription_id: str,
        *,
        product_id: str,
        proration_billing_mode: Literal["prorated_immediately", "full_immediately", "difference_immediately"],
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        on_payment_failure: Optional[Literal["prevent_change", "apply_change"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> None:
        """
        Args:
          product_id: Unique identifier of the product to subscribe to

          proration_billing_mode: Proration Billing Mode

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Addons for the new plan. Note : Leaving this empty would remove any existing
              addons

          discount_code: Optional discount code to apply to the new plan. If provided, validates and
              applies the discount to the plan change. If not provided and the subscription
              has an existing discount with `preserve_on_plan_change=true`, the existing
              discount will be preserved (if applicable to the new product).

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          on_payment_failure: Controls behavior when the plan change payment fails.

              - `prevent_change`: Keep subscription on current plan until payment succeeds
              - `apply_change` (default): Apply plan change immediately regardless of payment
                outcome

              If not specified, uses the business-level default setting.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        extra_headers = {"Accept": "*/*", **(extra_headers or {})}
        return await self._post(
            f"/subscriptions/{subscription_id}/change-plan",
            body=await async_maybe_transform(
                {
                    "product_id": product_id,
                    "proration_billing_mode": proration_billing_mode,
                    "quantity": quantity,
                    "addons": addons,
                    "discount_code": discount_code,
                    "metadata": metadata,
                    "on_payment_failure": on_payment_failure,
                },
                subscription_change_plan_params.SubscriptionChangePlanParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=NoneType,
        )

    async def charge(
        self,
        subscription_id: str,
        *,
        product_price: int,
        adaptive_currency_fees_inclusive: Optional[bool] | Omit = omit,
        customer_balance_config: Optional[subscription_charge_params.CustomerBalanceConfig] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        product_currency: Optional[Currency] | Omit = omit,
        product_description: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionChargeResponse:
        """Args:
          product_price: The product price.

        Represented in the lowest denomination of the currency (e.g.,
              cents for USD). For example, to charge $1.00, pass `100`.

          adaptive_currency_fees_inclusive: Whether adaptive currency fees should be included in the product_price (true) or
              added on top (false). This field is ignored if adaptive pricing is not enabled
              for the business.

          customer_balance_config: Specify how customer balance is used for the payment

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          product_currency: Optional currency of the product price. If not specified, defaults to the
              currency of the product.

          product_description: Optional product description override for billing and line items. If not
              specified, the stored description of the product will be used.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._post(
            f"/subscriptions/{subscription_id}/charge",
            body=await async_maybe_transform(
                {
                    "product_price": product_price,
                    "adaptive_currency_fees_inclusive": adaptive_currency_fees_inclusive,
                    "customer_balance_config": customer_balance_config,
                    "metadata": metadata,
                    "product_currency": product_currency,
                    "product_description": product_description,
                },
                subscription_charge_params.SubscriptionChargeParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionChargeResponse,
        )

    async def preview_change_plan(
        self,
        subscription_id: str,
        *,
        product_id: str,
        proration_billing_mode: Literal["prorated_immediately", "full_immediately", "difference_immediately"],
        quantity: int,
        addons: Optional[Iterable[AttachAddonParam]] | Omit = omit,
        discount_code: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        on_payment_failure: Optional[Literal["prevent_change", "apply_change"]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionPreviewChangePlanResponse:
        """
        Args:
          product_id: Unique identifier of the product to subscribe to

          proration_billing_mode: Proration Billing Mode

          quantity: Number of units to subscribe for. Must be at least 1.

          addons: Addons for the new plan. Note : Leaving this empty would remove any existing
              addons

          discount_code: Optional discount code to apply to the new plan. If provided, validates and
              applies the discount to the plan change. If not provided and the subscription
              has an existing discount with `preserve_on_plan_change=true`, the existing
              discount will be preserved (if applicable to the new product).

          metadata: Metadata for the payment. If not passed, the metadata of the subscription will
              be taken

          on_payment_failure: Controls behavior when the plan change payment fails.

              - `prevent_change`: Keep subscription on current plan until payment succeeds
              - `apply_change` (default): Apply plan change immediately regardless of payment
                outcome

              If not specified, uses the business-level default setting.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._post(
            f"/subscriptions/{subscription_id}/change-plan/preview",
            body=await async_maybe_transform(
                {
                    "product_id": product_id,
                    "proration_billing_mode": proration_billing_mode,
                    "quantity": quantity,
                    "addons": addons,
                    "discount_code": discount_code,
                    "metadata": metadata,
                    "on_payment_failure": on_payment_failure,
                },
                subscription_preview_change_plan_params.SubscriptionPreviewChangePlanParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionPreviewChangePlanResponse,
        )

    async def retrieve_credit_usage(
        self,
        subscription_id: str,
        *,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionRetrieveCreditUsageResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._get(
            f"/subscriptions/{subscription_id}/credit-usage",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionRetrieveCreditUsageResponse,
        )

    def retrieve_usage_history(
        self,
        subscription_id: str,
        *,
        end_date: Union[str, datetime, None] | Omit = omit,
        meter_id: Optional[str] | Omit = omit,
        page_number: Optional[int] | Omit = omit,
        page_size: Optional[int] | Omit = omit,
        start_date: Union[str, datetime, None] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[
        SubscriptionRetrieveUsageHistoryResponse,
        AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse],
    ]:
        """
        Get detailed usage history for a subscription that includes usage-based billing
        (metered components). This endpoint provides insights into customer usage
        patterns and billing calculations over time.

        ## What You'll Get:

        - **Billing periods**: Each item represents a billing cycle with start and end
          dates
        - **Meter usage**: Detailed breakdown of usage for each meter configured on the
          subscription
        - **Usage calculations**: Total units consumed, free threshold units, and
          chargeable units
        - **Historical tracking**: Complete audit trail of usage-based charges

        ## Use Cases:

        - **Customer support**: Investigate billing questions and usage discrepancies
        - **Usage analytics**: Analyze customer consumption patterns over time
        - **Billing transparency**: Provide customers with detailed usage breakdowns
        - **Revenue optimization**: Identify usage trends to optimize pricing strategies

        ## Filtering Options:

        - **Date range filtering**: Get usage history for specific time periods
        - **Meter-specific filtering**: Focus on usage for a particular meter
        - **Pagination**: Navigate through large usage histories efficiently

        ## Important Notes:

        - Only returns data for subscriptions with usage-based (metered) components
        - Usage history is organized by billing periods (subscription cycles)
        - Free threshold units are calculated and displayed separately from chargeable
          units
        - Historical data is preserved even if meter configurations change

        ## Example Query Patterns:

        - Get last 3 months:
          `?start_date=2024-01-01T00:00:00Z&end_date=2024-03-31T23:59:59Z`
        - Filter by meter: `?meter_id=mtr_api_requests`
        - Paginate results: `?page_size=20&page_number=1`
        - Recent usage: `?start_date=2024-03-01T00:00:00Z` (from March 1st to now)

        Args:
          end_date: Filter by end date (inclusive)

          meter_id: Filter by specific meter ID

          page_number: Page number (default: 0)

          page_size: Page size (default: 10, max: 100)

          start_date: Filter by start date (inclusive)

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return self._get_api_list(
            f"/subscriptions/{subscription_id}/usage-history",
            page=AsyncDefaultPageNumberPagination[SubscriptionRetrieveUsageHistoryResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "end_date": end_date,
                        "meter_id": meter_id,
                        "page_number": page_number,
                        "page_size": page_size,
                        "start_date": start_date,
                    },
                    subscription_retrieve_usage_history_params.SubscriptionRetrieveUsageHistoryParams,
                ),
            ),
            model=SubscriptionRetrieveUsageHistoryResponse,
        )

    @overload
    async def update_payment_method(
        self,
        subscription_id: str,
        *,
        type: Literal["new"],
        return_url: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        ...

    @overload
    async def update_payment_method(
        self,
        subscription_id: str,
        *,
        payment_method_id: str,
        type: Literal["existing"],
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        """
        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        ...

    @required_args(["type"], ["payment_method_id", "type"])
    async def update_payment_method(
        self,
        subscription_id: str,
        *,
        type: Literal["new"] | Literal["existing"],
        return_url: Optional[str] | Omit = omit,
        payment_method_id: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SubscriptionUpdatePaymentMethodResponse:
        if not subscription_id:
            raise ValueError(f"Expected a non-empty value for `subscription_id` but received {subscription_id!r}")
        return await self._post(
            f"/subscriptions/{subscription_id}/update-payment-method",
            body=await async_maybe_transform(
                {
                    "type": type,
                    "return_url": return_url,
                    "payment_method_id": payment_method_id,
                },
                subscription_update_payment_method_params.SubscriptionUpdatePaymentMethodParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=SubscriptionUpdatePaymentMethodResponse,
        )


class SubscriptionsResourceWithRawResponse:
    def __init__(self, subscriptions: SubscriptionsResource) -> None:
        self._subscriptions = subscriptions

        self.create = (  # pyright: ignore[reportDeprecated]
            to_raw_response_wrapper(
                subscriptions.create,  # pyright: ignore[reportDeprecated],
            )
        )
        self.retrieve = to_raw_response_wrapper(
            subscriptions.retrieve,
        )
        self.update = to_raw_response_wrapper(
            subscriptions.update,
        )
        self.list = to_raw_response_wrapper(
            subscriptions.list,
        )
        self.change_plan = to_raw_response_wrapper(
            subscriptions.change_plan,
        )
        self.charge = to_raw_response_wrapper(
            subscriptions.charge,
        )
        self.preview_change_plan = to_raw_response_wrapper(
            subscriptions.preview_change_plan,
        )
        self.retrieve_credit_usage = to_raw_response_wrapper(
            subscriptions.retrieve_credit_usage,
        )
        self.retrieve_usage_history = to_raw_response_wrapper(
            subscriptions.retrieve_usage_history,
        )
        self.update_payment_method = to_raw_response_wrapper(
            subscriptions.update_payment_method,
        )


class AsyncSubscriptionsResourceWithRawResponse:
    def __init__(self, subscriptions: AsyncSubscriptionsResource) -> None:
        self._subscriptions = subscriptions

        self.create = (  # pyright: ignore[reportDeprecated]
            async_to_raw_response_wrapper(
                subscriptions.create,  # pyright: ignore[reportDeprecated],
            )
        )
        self.retrieve = async_to_raw_response_wrapper(
            subscriptions.retrieve,
        )
        self.update = async_to_raw_response_wrapper(
            subscriptions.update,
        )
        self.list = async_to_raw_response_wrapper(
            subscriptions.list,
        )
        self.change_plan = async_to_raw_response_wrapper(
            subscriptions.change_plan,
        )
        self.charge = async_to_raw_response_wrapper(
            subscriptions.charge,
        )
        self.preview_change_plan = async_to_raw_response_wrapper(
            subscriptions.preview_change_plan,
        )
        self.retrieve_credit_usage = async_to_raw_response_wrapper(
            subscriptions.retrieve_credit_usage,
        )
        self.retrieve_usage_history = async_to_raw_response_wrapper(
            subscriptions.retrieve_usage_history,
        )
        self.update_payment_method = async_to_raw_response_wrapper(
            subscriptions.update_payment_method,
        )


class SubscriptionsResourceWithStreamingResponse:
    def __init__(self, subscriptions: SubscriptionsResource) -> None:
        self._subscriptions = subscriptions

        self.create = (  # pyright: ignore[reportDeprecated]
            to_streamed_response_wrapper(
                subscriptions.create,  # pyright: ignore[reportDeprecated],
            )
        )
        self.retrieve = to_streamed_response_wrapper(
            subscriptions.retrieve,
        )
        self.update = to_streamed_response_wrapper(
            subscriptions.update,
        )
        self.list = to_streamed_response_wrapper(
            subscriptions.list,
        )
        self.change_plan = to_streamed_response_wrapper(
            subscriptions.change_plan,
        )
        self.charge = to_streamed_response_wrapper(
            subscriptions.charge,
        )
        self.preview_change_plan = to_streamed_response_wrapper(
            subscriptions.preview_change_plan,
        )
        self.retrieve_credit_usage = to_streamed_response_wrapper(
            subscriptions.retrieve_credit_usage,
        )
        self.retrieve_usage_history = to_streamed_response_wrapper(
            subscriptions.retrieve_usage_history,
        )
        self.update_payment_method = to_streamed_response_wrapper(
            subscriptions.update_payment_method,
        )


class AsyncSubscriptionsResourceWithStreamingResponse:
    def __init__(self, subscriptions: AsyncSubscriptionsResource) -> None:
        self._subscriptions = subscriptions

        self.create = (  # pyright: ignore[reportDeprecated]
            async_to_streamed_response_wrapper(
                subscriptions.create,  # pyright: ignore[reportDeprecated],
            )
        )
        self.retrieve = async_to_streamed_response_wrapper(
            subscriptions.retrieve,
        )
        self.update = async_to_streamed_response_wrapper(
            subscriptions.update,
        )
        self.list = async_to_streamed_response_wrapper(
            subscriptions.list,
        )
        self.change_plan = async_to_streamed_response_wrapper(
            subscriptions.change_plan,
        )
        self.charge = async_to_streamed_response_wrapper(
            subscriptions.charge,
        )
        self.preview_change_plan = async_to_streamed_response_wrapper(
            subscriptions.preview_change_plan,
        )
        self.retrieve_credit_usage = async_to_streamed_response_wrapper(
            subscriptions.retrieve_credit_usage,
        )
        self.retrieve_usage_history = async_to_streamed_response_wrapper(
            subscriptions.retrieve_usage_history,
        )
        self.update_payment_method = async_to_streamed_response_wrapper(
            subscriptions.update_payment_method,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["PaymentMethodTypes"]

PaymentMethodTypes: TypeAlias = Literal[
    "ach",
    "affirm",
    "afterpay_clearpay",
    "alfamart",
    "ali_pay",
    "ali_pay_hk",
    "alma",
    "amazon_pay",
    "apple_pay",
    "atome",
    "bacs",
    "bancontact_card",
    "becs",
    "benefit",
    "bizum",
    "blik",
    "boleto",
    "bca_bank_transfer",
    "bni_va",
    "bri_va",
    "card_redirect",
    "cimb_va",
    "classic",
    "credit",
    "crypto_currency",
    "cashapp",
    "dana",
    "danamon_va",
    "debit",
    "duit_now",
    "efecty",
    "eft",
    "eps",
    "fps",
    "evoucher",
    "giropay",
    "givex",
    "google_pay",
    "go_pay",
    "gcash",
    "ideal",
    "interac",
    "indomaret",
    "klarna",
    "kakao_pay",
    "local_bank_redirect",
    "mandiri_va",
    "knet",
    "mb_way",
    "mobile_pay",
    "momo",
    "momo_atm",
    "multibanco",
    "online_banking_thailand",
    "online_banking_czech_republic",
    "online_banking_finland",
    "online_banking_fpx",
    "online_banking_poland",
    "online_banking_slovakia",
    "oxxo",
    "pago_efectivo",
    "permata_bank_transfer",
    "open_banking_uk",
    "pay_bright",
    "paypal",
    "paze",
    "pix",
    "pay_safe_card",
    "przelewy24",
    "prompt_pay",
    "pse",
    "red_compra",
    "red_pagos",
    "samsung_pay",
    "sepa",
    "sepa_bank_transfer",
    "sofort",
    "swish",
    "touch_n_go",
    "trustly",
    "twint",
    "upi_collect",
    "upi_intent",
    "vipps",
    "viet_qr",
    "venmo",
    "walley",
    "we_chat_pay",
    "seven_eleven",
    "lawson",
    "mini_stop",
    "family_mart",
    "seicomart",
    "pay_easy",
    "local_bank_transfer",
    "mifinity",
    "open_banking_pis",
    "direct_carrier_billing",
    "instant_bank_transfer",
    "billie",
    "zip",
    "revolut_pay",
    "naver_pay",
    "payco",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List

from ..._models import BaseModel

__all__ = ["HeaderRetrieveResponse"]


class HeaderRetrieveResponse(BaseModel):
    """The value of the headers is returned in the `headers` field.

    Sensitive headers that have been redacted are returned in the sensitive
    field.
    """

    headers: Dict[str, str]
    """List of headers configured"""

    sensitive: List[str]
    """Sensitive headers without the value"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, List, Optional
from typing_extensions import TypedDict

from .webhook_event_type import WebhookEventType

__all__ = ["WebhookUpdateParams"]


class WebhookUpdateParams(TypedDict, total=False):
    description: Optional[str]
    """Description of the webhook"""

    disabled: Optional[bool]
    """To Disable the endpoint, set it to true."""

    filter_types: Optional[List[WebhookEventType]]
    """Filter events to the endpoint.

    Webhook event will only be sent for events in the list.
    """

    metadata: Optional[Dict[str, str]]
    """Metadata"""

    rate_limit: Optional[int]
    """Rate limit"""

    url: Optional[str]
    """Url endpoint"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Union
from typing_extensions import TypeAlias

from .dispute_won_webhook_event import DisputeWonWebhookEvent
from .credit_added_webhook_event import CreditAddedWebhookEvent
from .dispute_lost_webhook_event import DisputeLostWebhookEvent
from .refund_failed_webhook_event import RefundFailedWebhookEvent
from .credit_expired_webhook_event import CreditExpiredWebhookEvent
from .dispute_opened_webhook_event import DisputeOpenedWebhookEvent
from .payment_failed_webhook_event import PaymentFailedWebhookEvent
from .credit_deducted_webhook_event import CreditDeductedWebhookEvent
from .dispute_expired_webhook_event import DisputeExpiredWebhookEvent
from .dispute_accepted_webhook_event import DisputeAcceptedWebhookEvent
from .refund_succeeded_webhook_event import RefundSucceededWebhookEvent
from .dispute_cancelled_webhook_event import DisputeCancelledWebhookEvent
from .payment_cancelled_webhook_event import PaymentCancelledWebhookEvent
from .payment_succeeded_webhook_event import PaymentSucceededWebhookEvent
from .credit_balance_low_webhook_event import CreditBalanceLowWebhookEvent
from .credit_rolled_over_webhook_event import CreditRolledOverWebhookEvent
from .dispute_challenged_webhook_event import DisputeChallengedWebhookEvent
from .payment_processing_webhook_event import PaymentProcessingWebhookEvent
from .license_key_created_webhook_event import LicenseKeyCreatedWebhookEvent
from .subscription_active_webhook_event import SubscriptionActiveWebhookEvent
from .subscription_failed_webhook_event import SubscriptionFailedWebhookEvent
from .subscription_expired_webhook_event import SubscriptionExpiredWebhookEvent
from .subscription_on_hold_webhook_event import SubscriptionOnHoldWebhookEvent
from .subscription_renewed_webhook_event import SubscriptionRenewedWebhookEvent
from .subscription_updated_webhook_event import SubscriptionUpdatedWebhookEvent
from .credit_overage_charged_webhook_event import CreditOverageChargedWebhookEvent
from .subscription_cancelled_webhook_event import SubscriptionCancelledWebhookEvent
from .credit_manual_adjustment_webhook_event import CreditManualAdjustmentWebhookEvent
from .credit_rollover_forfeited_webhook_event import CreditRolloverForfeitedWebhookEvent
from .subscription_plan_changed_webhook_event import SubscriptionPlanChangedWebhookEvent

__all__ = ["UnsafeUnwrapWebhookEvent"]

UnsafeUnwrapWebhookEvent: TypeAlias = Union[
    CreditAddedWebhookEvent,
    CreditBalanceLowWebhookEvent,
    CreditDeductedWebhookEvent,
    CreditExpiredWebhookEvent,
    CreditManualAdjustmentWebhookEvent,
    CreditOverageChargedWebhookEvent,
    CreditRolledOverWebhookEvent,
    CreditRolloverForfeitedWebhookEvent,
    DisputeAcceptedWebhookEvent,
    DisputeCancelledWebhookEvent,
    DisputeChallengedWebhookEvent,
    DisputeExpiredWebhookEvent,
    DisputeLostWebhookEvent,
    DisputeOpenedWebhookEvent,
    DisputeWonWebhookEvent,
    LicenseKeyCreatedWebhookEvent,
    PaymentCancelledWebhookEvent,
    PaymentFailedWebhookEvent,
    PaymentProcessingWebhookEvent,
    PaymentSucceededWebhookEvent,
    RefundFailedWebhookEvent,
    RefundSucceededWebhookEvent,
    SubscriptionActiveWebhookEvent,
    SubscriptionCancelledWebhookEvent,
    SubscriptionExpiredWebhookEvent,
    SubscriptionFailedWebhookEvent,
    SubscriptionOnHoldWebhookEvent,
    SubscriptionPlanChangedWebhookEvent,
    SubscriptionRenewedWebhookEvent,
    SubscriptionUpdatedWebhookEvent,
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import Literal

import httpx

from ....types import Currency
from ...._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ...._utils import maybe_transform, async_maybe_transform
from ...._compat import cached_property
from ...._resource import SyncAPIResource, AsyncAPIResource
from ...._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ....pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination
from ...._base_client import AsyncPaginator, make_request_options
from ....types.currency import Currency
from ....types.customers.wallets import ledger_entry_list_params, ledger_entry_create_params
from ....types.customers.customer_wallet import CustomerWallet
from ....types.customers.wallets.customer_wallet_transaction import CustomerWalletTransaction

__all__ = ["LedgerEntriesResource", "AsyncLedgerEntriesResource"]


class LedgerEntriesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> LedgerEntriesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return LedgerEntriesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> LedgerEntriesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return LedgerEntriesResourceWithStreamingResponse(self)

    def create(
        self,
        customer_id: str,
        *,
        amount: int,
        currency: Currency,
        entry_type: Literal["credit", "debit"],
        idempotency_key: Optional[str] | Omit = omit,
        reason: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerWallet:
        """
        Args:
          currency: Currency of the wallet to adjust

          entry_type: Type of ledger entry - credit or debit

          idempotency_key: Optional idempotency key to prevent duplicate entries

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._post(
            f"/customers/{customer_id}/wallets/ledger-entries",
            body=maybe_transform(
                {
                    "amount": amount,
                    "currency": currency,
                    "entry_type": entry_type,
                    "idempotency_key": idempotency_key,
                    "reason": reason,
                },
                ledger_entry_create_params.LedgerEntryCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerWallet,
        )

    def list(
        self,
        customer_id: str,
        *,
        currency: Currency | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[CustomerWalletTransaction]:
        """
        Args:
          currency: Optional currency filter

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/customers/{customer_id}/wallets/ledger-entries",
            page=SyncDefaultPageNumberPagination[CustomerWalletTransaction],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "currency": currency,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    ledger_entry_list_params.LedgerEntryListParams,
                ),
            ),
            model=CustomerWalletTransaction,
        )


class AsyncLedgerEntriesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncLedgerEntriesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncLedgerEntriesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncLedgerEntriesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncLedgerEntriesResourceWithStreamingResponse(self)

    async def create(
        self,
        customer_id: str,
        *,
        amount: int,
        currency: Currency,
        entry_type: Literal["credit", "debit"],
        idempotency_key: Optional[str] | Omit = omit,
        reason: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerWallet:
        """
        Args:
          currency: Currency of the wallet to adjust

          entry_type: Type of ledger entry - credit or debit

          idempotency_key: Optional idempotency key to prevent duplicate entries

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._post(
            f"/customers/{customer_id}/wallets/ledger-entries",
            body=await async_maybe_transform(
                {
                    "amount": amount,
                    "currency": currency,
                    "entry_type": entry_type,
                    "idempotency_key": idempotency_key,
                    "reason": reason,
                },
                ledger_entry_create_params.LedgerEntryCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerWallet,
        )

    def list(
        self,
        customer_id: str,
        *,
        currency: Currency | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[CustomerWalletTransaction, AsyncDefaultPageNumberPagination[CustomerWalletTransaction]]:
        """
        Args:
          currency: Optional currency filter

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/customers/{customer_id}/wallets/ledger-entries",
            page=AsyncDefaultPageNumberPagination[CustomerWalletTransaction],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "currency": currency,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    ledger_entry_list_params.LedgerEntryListParams,
                ),
            ),
            model=CustomerWalletTransaction,
        )


class LedgerEntriesResourceWithRawResponse:
    def __init__(self, ledger_entries: LedgerEntriesResource) -> None:
        self._ledger_entries = ledger_entries

        self.create = to_raw_response_wrapper(
            ledger_entries.create,
        )
        self.list = to_raw_response_wrapper(
            ledger_entries.list,
        )


class AsyncLedgerEntriesResourceWithRawResponse:
    def __init__(self, ledger_entries: AsyncLedgerEntriesResource) -> None:
        self._ledger_entries = ledger_entries

        self.create = async_to_raw_response_wrapper(
            ledger_entries.create,
        )
        self.list = async_to_raw_response_wrapper(
            ledger_entries.list,
        )


class LedgerEntriesResourceWithStreamingResponse:
    def __init__(self, ledger_entries: LedgerEntriesResource) -> None:
        self._ledger_entries = ledger_entries

        self.create = to_streamed_response_wrapper(
            ledger_entries.create,
        )
        self.list = to_streamed_response_wrapper(
            ledger_entries.list,
        )


class AsyncLedgerEntriesResourceWithStreamingResponse:
    def __init__(self, ledger_entries: AsyncLedgerEntriesResource) -> None:
        self._ledger_entries = ledger_entries

        self.create = async_to_streamed_response_wrapper(
            ledger_entries.create,
        )
        self.list = async_to_streamed_response_wrapper(
            ledger_entries.list,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from .._models import BaseModel
from .intent_status import IntentStatus

__all__ = ["CheckoutSessionStatus"]


class CheckoutSessionStatus(BaseModel):
    id: str
    """Id of the checkout session"""

    created_at: datetime
    """Created at timestamp"""

    customer_email: Optional[str] = None
    """Customer email: prefers payment's customer, falls back to session"""

    customer_name: Optional[str] = None
    """Customer name: prefers payment's customer, falls back to session"""

    payment_id: Optional[str] = None
    """Id of the payment created by the checkout sessions.

    Null if checkout sessions is still at the details collection stage.
    """

    payment_status: Optional[IntentStatus] = None
    """status of the payment.

    Null if checkout sessions is still at the details collection stage.
    """



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime

from .._models import BaseModel
from .addon_cart_response_item import AddonCartResponseItem
from .customer_limited_details import CustomerLimitedDetails
from .one_time_product_cart_item import OneTimeProductCartItem

__all__ = ["SubscriptionCreateResponse"]


class SubscriptionCreateResponse(BaseModel):
    addons: List[AddonCartResponseItem]
    """Addons associated with this subscription"""

    customer: CustomerLimitedDetails
    """Customer details associated with this subscription"""

    metadata: Dict[str, str]
    """Additional metadata associated with the subscription"""

    payment_id: str
    """First payment id for the subscription"""

    recurring_pre_tax_amount: int
    """
    Tax will be added to the amount and charged to the customer on each billing
    cycle
    """

    subscription_id: str
    """Unique identifier for the subscription"""

    client_secret: Optional[str] = None
    """
    Client secret used to load Dodo checkout SDK NOTE : Dodo checkout SDK will be
    coming soon
    """

    discount_id: Optional[str] = None
    """The discount id if discount is applied"""

    expires_on: Optional[datetime] = None
    """Expiry timestamp of the payment link"""

    one_time_product_cart: Optional[List[OneTimeProductCartItem]] = None
    """One time products associated with the purchase of subscription"""

    payment_link: Optional[str] = None
    """URL to checkout page"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional

import httpx

from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination
from ..._base_client import AsyncPaginator, make_request_options
from ...types.products import short_link_list_params, short_link_create_params
from ...types.products.short_link_list_response import ShortLinkListResponse
from ...types.products.short_link_create_response import ShortLinkCreateResponse

__all__ = ["ShortLinksResource", "AsyncShortLinksResource"]


class ShortLinksResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> ShortLinksResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return ShortLinksResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> ShortLinksResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return ShortLinksResourceWithStreamingResponse(self)

    def create(
        self,
        id: str,
        *,
        slug: str,
        static_checkout_params: Optional[Dict[str, str]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ShortLinkCreateResponse:
        """Gives a Short Checkout URL with custom slug for a product.

        Uses a Static
        Checkout URL under the hood.

        Args:
          slug: Slug for the short link.

          static_checkout_params: Static Checkout URL parameters to apply to the resulting short URL.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return self._post(
            f"/products/{id}/short_links",
            body=maybe_transform(
                {
                    "slug": slug,
                    "static_checkout_params": static_checkout_params,
                },
                short_link_create_params.ShortLinkCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ShortLinkCreateResponse,
        )

    def list(
        self,
        *,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        product_id: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[ShortLinkListResponse]:
        """
        Lists all short links created by the business.

        Args:
          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          product_id: Filter by product ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/products/short_links",
            page=SyncDefaultPageNumberPagination[ShortLinkListResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "page_number": page_number,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    short_link_list_params.ShortLinkListParams,
                ),
            ),
            model=ShortLinkListResponse,
        )


class AsyncShortLinksResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncShortLinksResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncShortLinksResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncShortLinksResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncShortLinksResourceWithStreamingResponse(self)

    async def create(
        self,
        id: str,
        *,
        slug: str,
        static_checkout_params: Optional[Dict[str, str]] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> ShortLinkCreateResponse:
        """Gives a Short Checkout URL with custom slug for a product.

        Uses a Static
        Checkout URL under the hood.

        Args:
          slug: Slug for the short link.

          static_checkout_params: Static Checkout URL parameters to apply to the resulting short URL.

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not id:
            raise ValueError(f"Expected a non-empty value for `id` but received {id!r}")
        return await self._post(
            f"/products/{id}/short_links",
            body=await async_maybe_transform(
                {
                    "slug": slug,
                    "static_checkout_params": static_checkout_params,
                },
                short_link_create_params.ShortLinkCreateParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=ShortLinkCreateResponse,
        )

    def list(
        self,
        *,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        product_id: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[ShortLinkListResponse, AsyncDefaultPageNumberPagination[ShortLinkListResponse]]:
        """
        Lists all short links created by the business.

        Args:
          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          product_id: Filter by product ID

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        return self._get_api_list(
            "/products/short_links",
            page=AsyncDefaultPageNumberPagination[ShortLinkListResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "page_number": page_number,
                        "page_size": page_size,
                        "product_id": product_id,
                    },
                    short_link_list_params.ShortLinkListParams,
                ),
            ),
            model=ShortLinkListResponse,
        )


class ShortLinksResourceWithRawResponse:
    def __init__(self, short_links: ShortLinksResource) -> None:
        self._short_links = short_links

        self.create = to_raw_response_wrapper(
            short_links.create,
        )
        self.list = to_raw_response_wrapper(
            short_links.list,
        )


class AsyncShortLinksResourceWithRawResponse:
    def __init__(self, short_links: AsyncShortLinksResource) -> None:
        self._short_links = short_links

        self.create = async_to_raw_response_wrapper(
            short_links.create,
        )
        self.list = async_to_raw_response_wrapper(
            short_links.list,
        )


class ShortLinksResourceWithStreamingResponse:
    def __init__(self, short_links: ShortLinksResource) -> None:
        self._short_links = short_links

        self.create = to_streamed_response_wrapper(
            short_links.create,
        )
        self.list = to_streamed_response_wrapper(
            short_links.list,
        )


class AsyncShortLinksResourceWithStreamingResponse:
    def __init__(self, short_links: AsyncShortLinksResource) -> None:
        self._short_links = short_links

        self.create = async_to_streamed_response_wrapper(
            short_links.create,
        )
        self.list = async_to_streamed_response_wrapper(
            short_links.list,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import List, Optional

from .._models import BaseModel
from .digital_product_delivery_file import DigitalProductDeliveryFile

__all__ = ["DigitalProductDelivery"]


class DigitalProductDelivery(BaseModel):
    external_url: Optional[str] = None
    """External URL to digital product"""

    files: Optional[List[DigitalProductDeliveryFile]] = None
    """Uploaded files ids of digital product"""

    instructions: Optional[str] = None
    """Instructions to download and use the digital product"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .customer_wallet import CustomerWallet as CustomerWallet
from .wallet_list_response import WalletListResponse as WalletListResponse
from .customer_portal_create_params import CustomerPortalCreateParams as CustomerPortalCreateParams



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .images import (
    ImagesResource,
    AsyncImagesResource,
    ImagesResourceWithRawResponse,
    AsyncImagesResourceWithRawResponse,
    ImagesResourceWithStreamingResponse,
    AsyncImagesResourceWithStreamingResponse,
)
from .products import (
    ProductsResource,
    AsyncProductsResource,
    ProductsResourceWithRawResponse,
    AsyncProductsResourceWithRawResponse,
    ProductsResourceWithStreamingResponse,
    AsyncProductsResourceWithStreamingResponse,
)
from .short_links import (
    ShortLinksResource,
    AsyncShortLinksResource,
    ShortLinksResourceWithRawResponse,
    AsyncShortLinksResourceWithRawResponse,
    ShortLinksResourceWithStreamingResponse,
    AsyncShortLinksResourceWithStreamingResponse,
)

__all__ = [
    "ImagesResource",
    "AsyncImagesResource",
    "ImagesResourceWithRawResponse",
    "AsyncImagesResourceWithRawResponse",
    "ImagesResourceWithStreamingResponse",
    "AsyncImagesResourceWithStreamingResponse",
    "ShortLinksResource",
    "AsyncShortLinksResource",
    "ShortLinksResourceWithRawResponse",
    "AsyncShortLinksResourceWithRawResponse",
    "ShortLinksResourceWithStreamingResponse",
    "AsyncShortLinksResourceWithStreamingResponse",
    "ProductsResource",
    "AsyncProductsResource",
    "ProductsResourceWithRawResponse",
    "AsyncProductsResourceWithRawResponse",
    "ProductsResourceWithStreamingResponse",
    "AsyncProductsResourceWithStreamingResponse",
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, Optional
from datetime import datetime
from typing_extensions import Literal

from ..._models import BaseModel

__all__ = ["BalanceListGrantsResponse"]


class BalanceListGrantsResponse(BaseModel):
    """Response for a credit grant"""

    id: str

    created_at: datetime

    credit_entitlement_id: str

    customer_id: str

    initial_amount: str

    is_expired: bool

    is_rolled_over: bool

    remaining_amount: str

    rollover_count: int

    source_type: Literal["subscription", "one_time", "addon", "api", "rollover"]

    updated_at: datetime

    expires_at: Optional[datetime] = None

    metadata: Optional[Dict[str, str]] = None

    parent_grant_id: Optional[str] = None

    source_id: Optional[str] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["CbbOverageBehavior"]

CbbOverageBehavior: TypeAlias = Literal[
    "forgive_at_reset", "invoice_at_billing", "carry_deficit", "carry_deficit_auto_repay"
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from .._models import BaseModel

__all__ = ["MeterAggregation"]


class MeterAggregation(BaseModel):
    type: Literal["count", "sum", "max", "last"]
    """Aggregation type for the meter"""

    key: Optional[str] = None
    """Required when type is not COUNT"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from datetime import datetime

from ..._models import BaseModel

__all__ = ["ShortLinkListResponse"]


class ShortLinkListResponse(BaseModel):
    created_at: datetime
    """When the short url was created"""

    full_url: str
    """Full URL the short url redirects to"""

    product_id: str
    """Product ID associated with the short link"""

    short_url: str
    """Short URL"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from datetime import datetime

from ..._models import BaseModel

__all__ = ["CustomerCreditBalance"]


class CustomerCreditBalance(BaseModel):
    """Response for a customer's credit balance"""

    id: str

    balance: str

    created_at: datetime

    credit_entitlement_id: str

    customer_id: str

    overage: str

    updated_at: datetime

    last_transaction_at: Optional[datetime] = None



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Optional
from typing_extensions import TypedDict

__all__ = ["BrandCreateParams"]


class BrandCreateParams(TypedDict, total=False):
    description: Optional[str]

    name: Optional[str]

    statement_descriptor: Optional[str]

    support_email: Optional[str]

    url: Optional[str]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Union, Optional
from datetime import datetime
from typing_extensions import Literal

import httpx

from ..._types import Body, Omit, Query, Headers, NotGiven, omit, not_given
from ..._utils import maybe_transform, async_maybe_transform
from ..._compat import cached_property
from ..._resource import SyncAPIResource, AsyncAPIResource
from ..._response import (
    to_raw_response_wrapper,
    to_streamed_response_wrapper,
    async_to_raw_response_wrapper,
    async_to_streamed_response_wrapper,
)
from ...pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination
from ..._base_client import AsyncPaginator, make_request_options
from ...types.credit_entitlements import (
    LedgerEntryType,
    balance_list_params,
    balance_list_grants_params,
    balance_list_ledger_params,
    balance_create_ledger_entry_params,
)
from ...types.credit_entitlements.ledger_entry_type import LedgerEntryType
from ...types.credit_entitlements.credit_ledger_entry import CreditLedgerEntry
from ...types.credit_entitlements.customer_credit_balance import CustomerCreditBalance
from ...types.credit_entitlements.balance_list_grants_response import BalanceListGrantsResponse
from ...types.credit_entitlements.balance_create_ledger_entry_response import BalanceCreateLedgerEntryResponse

__all__ = ["BalancesResource", "AsyncBalancesResource"]


class BalancesResource(SyncAPIResource):
    @cached_property
    def with_raw_response(self) -> BalancesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return BalancesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> BalancesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return BalancesResourceWithStreamingResponse(self)

    def retrieve(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerCreditBalance:
        """
        Returns the credit balance details for a specific customer and credit
        entitlement.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Responses

        - `200 OK` - Returns the customer's balance
        - `404 Not Found` - Credit entitlement or customer balance not found
        - `500 Internal Server Error` - Database or server error

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerCreditBalance,
        )

    def list(
        self,
        credit_entitlement_id: str,
        *,
        customer_id: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[CustomerCreditBalance]:
        """
        Returns a paginated list of customer credit balances for the given credit
        entitlement.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `customer_id` - Optional filter by specific customer

        # Responses

        - `200 OK` - Returns list of customer balances
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          customer_id: Filter by specific customer ID

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances",
            page=SyncDefaultPageNumberPagination[CustomerCreditBalance],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "customer_id": customer_id,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    balance_list_params.BalanceListParams,
                ),
            ),
            model=CustomerCreditBalance,
        )

    def create_ledger_entry(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        amount: str,
        entry_type: LedgerEntryType,
        expires_at: Union[str, datetime, None] | Omit = omit,
        idempotency_key: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        reason: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BalanceCreateLedgerEntryResponse:
        """For credit entries, a new grant is created.

        For debit entries, credits are
        deducted from existing grants using FIFO (oldest first).

        # Authentication

        Requires an API key with `Editor` role.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Request Body

        - `entry_type` - "credit" or "debit"
        - `amount` - Amount to credit or debit
        - `reason` - Optional human-readable reason
        - `expires_at` - Optional expiration for credited amount (only for credit type)
        - `idempotency_key` - Optional key to prevent duplicate entries

        # Responses

        - `201 Created` - Ledger entry created successfully
        - `400 Bad Request` - Invalid request (e.g., debit with insufficient balance)
        - `404 Not Found` - Credit entitlement or customer not found
        - `409 Conflict` - Idempotency key already exists
        - `500 Internal Server Error` - Database or server error

        Args:
          amount: Amount to credit or debit

          entry_type: Entry type: credit or debit

          expires_at: Expiration for credited amount (only for credit type)

          idempotency_key: Idempotency key to prevent duplicate entries

          metadata: Optional metadata (max 50 key-value pairs, key max 40 chars, value max 500
              chars)

          reason: Human-readable reason for the entry

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._post(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/ledger-entries",
            body=maybe_transform(
                {
                    "amount": amount,
                    "entry_type": entry_type,
                    "expires_at": expires_at,
                    "idempotency_key": idempotency_key,
                    "metadata": metadata,
                    "reason": reason,
                },
                balance_create_ledger_entry_params.BalanceCreateLedgerEntryParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BalanceCreateLedgerEntryResponse,
        )

    def list_grants(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        status: Literal["active", "expired", "depleted"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[BalanceListGrantsResponse]:
        """
        Returns a paginated list of credit grants with optional filtering by status.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `status` - Filter by status: active, expired, depleted

        # Responses

        - `200 OK` - Returns list of grants
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          status: Filter by grant status: active, expired, depleted

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/grants",
            page=SyncDefaultPageNumberPagination[BalanceListGrantsResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "page_number": page_number,
                        "page_size": page_size,
                        "status": status,
                    },
                    balance_list_grants_params.BalanceListGrantsParams,
                ),
            ),
            model=BalanceListGrantsResponse,
        )

    def list_ledger(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        end_date: Union[str, datetime] | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        start_date: Union[str, datetime] | Omit = omit,
        transaction_type: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> SyncDefaultPageNumberPagination[CreditLedgerEntry]:
        """
        Returns a paginated list of credit transaction history with optional filtering.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `transaction_type` - Filter by transaction type
        - `start_date` - Filter entries from this date
        - `end_date` - Filter entries until this date

        # Responses

        - `200 OK` - Returns list of ledger entries
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          end_date: Filter by end date

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          start_date: Filter by start date

          transaction_type: Filter by transaction type (snake_case: credit_added, credit_deducted,
              credit_expired, etc.)

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/ledger",
            page=SyncDefaultPageNumberPagination[CreditLedgerEntry],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "end_date": end_date,
                        "page_number": page_number,
                        "page_size": page_size,
                        "start_date": start_date,
                        "transaction_type": transaction_type,
                    },
                    balance_list_ledger_params.BalanceListLedgerParams,
                ),
            ),
            model=CreditLedgerEntry,
        )


class AsyncBalancesResource(AsyncAPIResource):
    @cached_property
    def with_raw_response(self) -> AsyncBalancesResourceWithRawResponse:
        """
        This property can be used as a prefix for any HTTP method call to return
        the raw response object instead of the parsed content.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#accessing-raw-response-data-eg-headers
        """
        return AsyncBalancesResourceWithRawResponse(self)

    @cached_property
    def with_streaming_response(self) -> AsyncBalancesResourceWithStreamingResponse:
        """
        An alternative to `.with_raw_response` that doesn't eagerly read the response body.

        For more information, see https://www.github.com/dodopayments/dodopayments-python#with_streaming_response
        """
        return AsyncBalancesResourceWithStreamingResponse(self)

    async def retrieve(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> CustomerCreditBalance:
        """
        Returns the credit balance details for a specific customer and credit
        entitlement.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Responses

        - `200 OK` - Returns the customer's balance
        - `404 Not Found` - Credit entitlement or customer balance not found
        - `500 Internal Server Error` - Database or server error

        Args:
          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._get(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}",
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=CustomerCreditBalance,
        )

    def list(
        self,
        credit_entitlement_id: str,
        *,
        customer_id: str | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[CustomerCreditBalance, AsyncDefaultPageNumberPagination[CustomerCreditBalance]]:
        """
        Returns a paginated list of customer credit balances for the given credit
        entitlement.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `customer_id` - Optional filter by specific customer

        # Responses

        - `200 OK` - Returns list of customer balances
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          customer_id: Filter by specific customer ID

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances",
            page=AsyncDefaultPageNumberPagination[CustomerCreditBalance],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "customer_id": customer_id,
                        "page_number": page_number,
                        "page_size": page_size,
                    },
                    balance_list_params.BalanceListParams,
                ),
            ),
            model=CustomerCreditBalance,
        )

    async def create_ledger_entry(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        amount: str,
        entry_type: LedgerEntryType,
        expires_at: Union[str, datetime, None] | Omit = omit,
        idempotency_key: Optional[str] | Omit = omit,
        metadata: Optional[Dict[str, str]] | Omit = omit,
        reason: Optional[str] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> BalanceCreateLedgerEntryResponse:
        """For credit entries, a new grant is created.

        For debit entries, credits are
        deducted from existing grants using FIFO (oldest first).

        # Authentication

        Requires an API key with `Editor` role.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Request Body

        - `entry_type` - "credit" or "debit"
        - `amount` - Amount to credit or debit
        - `reason` - Optional human-readable reason
        - `expires_at` - Optional expiration for credited amount (only for credit type)
        - `idempotency_key` - Optional key to prevent duplicate entries

        # Responses

        - `201 Created` - Ledger entry created successfully
        - `400 Bad Request` - Invalid request (e.g., debit with insufficient balance)
        - `404 Not Found` - Credit entitlement or customer not found
        - `409 Conflict` - Idempotency key already exists
        - `500 Internal Server Error` - Database or server error

        Args:
          amount: Amount to credit or debit

          entry_type: Entry type: credit or debit

          expires_at: Expiration for credited amount (only for credit type)

          idempotency_key: Idempotency key to prevent duplicate entries

          metadata: Optional metadata (max 50 key-value pairs, key max 40 chars, value max 500
              chars)

          reason: Human-readable reason for the entry

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return await self._post(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/ledger-entries",
            body=await async_maybe_transform(
                {
                    "amount": amount,
                    "entry_type": entry_type,
                    "expires_at": expires_at,
                    "idempotency_key": idempotency_key,
                    "metadata": metadata,
                    "reason": reason,
                },
                balance_create_ledger_entry_params.BalanceCreateLedgerEntryParams,
            ),
            options=make_request_options(
                extra_headers=extra_headers, extra_query=extra_query, extra_body=extra_body, timeout=timeout
            ),
            cast_to=BalanceCreateLedgerEntryResponse,
        )

    def list_grants(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        status: Literal["active", "expired", "depleted"] | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[BalanceListGrantsResponse, AsyncDefaultPageNumberPagination[BalanceListGrantsResponse]]:
        """
        Returns a paginated list of credit grants with optional filtering by status.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `status` - Filter by status: active, expired, depleted

        # Responses

        - `200 OK` - Returns list of grants
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          status: Filter by grant status: active, expired, depleted

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/grants",
            page=AsyncDefaultPageNumberPagination[BalanceListGrantsResponse],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "page_number": page_number,
                        "page_size": page_size,
                        "status": status,
                    },
                    balance_list_grants_params.BalanceListGrantsParams,
                ),
            ),
            model=BalanceListGrantsResponse,
        )

    def list_ledger(
        self,
        customer_id: str,
        *,
        credit_entitlement_id: str,
        end_date: Union[str, datetime] | Omit = omit,
        page_number: int | Omit = omit,
        page_size: int | Omit = omit,
        start_date: Union[str, datetime] | Omit = omit,
        transaction_type: str | Omit = omit,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = not_given,
    ) -> AsyncPaginator[CreditLedgerEntry, AsyncDefaultPageNumberPagination[CreditLedgerEntry]]:
        """
        Returns a paginated list of credit transaction history with optional filtering.

        # Authentication

        Requires an API key with `Viewer` role or higher.

        # Path Parameters

        - `credit_entitlement_id` - The unique identifier of the credit entitlement
        - `customer_id` - The unique identifier of the customer

        # Query Parameters

        - `page_size` - Number of items per page (default: 10, max: 100)
        - `page_number` - Zero-based page number (default: 0)
        - `transaction_type` - Filter by transaction type
        - `start_date` - Filter entries from this date
        - `end_date` - Filter entries until this date

        # Responses

        - `200 OK` - Returns list of ledger entries
        - `404 Not Found` - Credit entitlement not found
        - `500 Internal Server Error` - Database or server error

        Args:
          end_date: Filter by end date

          page_number: Page number default is 0

          page_size: Page size default is 10 max is 100

          start_date: Filter by start date

          transaction_type: Filter by transaction type (snake_case: credit_added, credit_deducted,
              credit_expired, etc.)

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        if not credit_entitlement_id:
            raise ValueError(
                f"Expected a non-empty value for `credit_entitlement_id` but received {credit_entitlement_id!r}"
            )
        if not customer_id:
            raise ValueError(f"Expected a non-empty value for `customer_id` but received {customer_id!r}")
        return self._get_api_list(
            f"/credit-entitlements/{credit_entitlement_id}/balances/{customer_id}/ledger",
            page=AsyncDefaultPageNumberPagination[CreditLedgerEntry],
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
                query=maybe_transform(
                    {
                        "end_date": end_date,
                        "page_number": page_number,
                        "page_size": page_size,
                        "start_date": start_date,
                        "transaction_type": transaction_type,
                    },
                    balance_list_ledger_params.BalanceListLedgerParams,
                ),
            ),
            model=CreditLedgerEntry,
        )


class BalancesResourceWithRawResponse:
    def __init__(self, balances: BalancesResource) -> None:
        self._balances = balances

        self.retrieve = to_raw_response_wrapper(
            balances.retrieve,
        )
        self.list = to_raw_response_wrapper(
            balances.list,
        )
        self.create_ledger_entry = to_raw_response_wrapper(
            balances.create_ledger_entry,
        )
        self.list_grants = to_raw_response_wrapper(
            balances.list_grants,
        )
        self.list_ledger = to_raw_response_wrapper(
            balances.list_ledger,
        )


class AsyncBalancesResourceWithRawResponse:
    def __init__(self, balances: AsyncBalancesResource) -> None:
        self._balances = balances

        self.retrieve = async_to_raw_response_wrapper(
            balances.retrieve,
        )
        self.list = async_to_raw_response_wrapper(
            balances.list,
        )
        self.create_ledger_entry = async_to_raw_response_wrapper(
            balances.create_ledger_entry,
        )
        self.list_grants = async_to_raw_response_wrapper(
            balances.list_grants,
        )
        self.list_ledger = async_to_raw_response_wrapper(
            balances.list_ledger,
        )


class BalancesResourceWithStreamingResponse:
    def __init__(self, balances: BalancesResource) -> None:
        self._balances = balances

        self.retrieve = to_streamed_response_wrapper(
            balances.retrieve,
        )
        self.list = to_streamed_response_wrapper(
            balances.list,
        )
        self.create_ledger_entry = to_streamed_response_wrapper(
            balances.create_ledger_entry,
        )
        self.list_grants = to_streamed_response_wrapper(
            balances.list_grants,
        )
        self.list_ledger = to_streamed_response_wrapper(
            balances.list_ledger,
        )


class AsyncBalancesResourceWithStreamingResponse:
    def __init__(self, balances: AsyncBalancesResource) -> None:
        self._balances = balances

        self.retrieve = async_to_streamed_response_wrapper(
            balances.retrieve,
        )
        self.list = async_to_streamed_response_wrapper(
            balances.list,
        )
        self.create_ledger_entry = async_to_streamed_response_wrapper(
            balances.create_ledger_entry,
        )
        self.list_grants = async_to_streamed_response_wrapper(
            balances.list_grants,
        )
        self.list_ledger = async_to_streamed_response_wrapper(
            balances.list_ledger,
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Dict, List, Optional
from datetime import datetime

from .price import Price
from .._models import BaseModel
from .tax_category import TaxCategory
from .license_key_duration import LicenseKeyDuration
from .digital_product_delivery import DigitalProductDelivery
from .credit_entitlement_mapping_response import CreditEntitlementMappingResponse

__all__ = ["Product"]


class Product(BaseModel):
    brand_id: str

    business_id: str
    """Unique identifier for the business to which the product belongs."""

    created_at: datetime
    """Timestamp when the product was created."""

    credit_entitlements: List[CreditEntitlementMappingResponse]
    """Attached credit entitlements with settings"""

    is_recurring: bool
    """Indicates if the product is recurring (e.g., subscriptions)."""

    license_key_enabled: bool
    """Indicates whether the product requires a license key."""

    metadata: Dict[str, str]
    """Additional custom data associated with the product"""

    price: Price
    """Pricing information for the product."""

    product_id: str
    """Unique identifier for the product."""

    tax_category: TaxCategory
    """Tax category associated with the product."""

    updated_at: datetime
    """Timestamp when the product was last updated."""

    addons: Optional[List[str]] = None
    """Available Addons for subscription products"""

    description: Optional[str] = None
    """Description of the product, optional."""

    digital_product_delivery: Optional[DigitalProductDelivery] = None

    image: Optional[str] = None
    """URL of the product image, optional."""

    license_key_activation_message: Optional[str] = None
    """Message sent upon license key activation, if applicable."""

    license_key_activations_limit: Optional[int] = None
    """Limit on the number of activations for the license key, if enabled."""

    license_key_duration: Optional[LicenseKeyDuration] = None
    """Duration of the license key validity, if enabled."""

    name: Optional[str] = None
    """Name of the product, optional."""

    product_collection_id: Optional[str] = None
    """The product collection ID this product belongs to, if any"""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

import os
from typing import Any, cast

import pytest

from tests.utils import assert_matches_type
from dodopayments import DodoPayments, AsyncDodoPayments
from dodopayments.types import LicenseKey
from dodopayments._utils import parse_datetime
from dodopayments.pagination import SyncDefaultPageNumberPagination, AsyncDefaultPageNumberPagination

base_url = os.environ.get("TEST_API_BASE_URL", "http://127.0.0.1:4010")


class TestLicenseKeys:
    parametrize = pytest.mark.parametrize("client", [False, True], indirect=True, ids=["loose", "strict"])

    @parametrize
    def test_method_retrieve(self, client: DodoPayments) -> None:
        license_key = client.license_keys.retrieve(
            "lic_123",
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    def test_raw_response_retrieve(self, client: DodoPayments) -> None:
        response = client.license_keys.with_raw_response.retrieve(
            "lic_123",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = response.parse()
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    def test_streaming_response_retrieve(self, client: DodoPayments) -> None:
        with client.license_keys.with_streaming_response.retrieve(
            "lic_123",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = response.parse()
            assert_matches_type(LicenseKey, license_key, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_retrieve(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            client.license_keys.with_raw_response.retrieve(
                "",
            )

    @parametrize
    def test_method_update(self, client: DodoPayments) -> None:
        license_key = client.license_keys.update(
            id="lic_123",
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    def test_method_update_with_all_params(self, client: DodoPayments) -> None:
        license_key = client.license_keys.update(
            id="lic_123",
            activations_limit=0,
            disabled=True,
            expires_at=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    def test_raw_response_update(self, client: DodoPayments) -> None:
        response = client.license_keys.with_raw_response.update(
            id="lic_123",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = response.parse()
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    def test_streaming_response_update(self, client: DodoPayments) -> None:
        with client.license_keys.with_streaming_response.update(
            id="lic_123",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = response.parse()
            assert_matches_type(LicenseKey, license_key, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    def test_path_params_update(self, client: DodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            client.license_keys.with_raw_response.update(
                id="",
            )

    @parametrize
    def test_method_list(self, client: DodoPayments) -> None:
        license_key = client.license_keys.list()
        assert_matches_type(SyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    def test_method_list_with_all_params(self, client: DodoPayments) -> None:
        license_key = client.license_keys.list(
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            customer_id="customer_id",
            page_number=0,
            page_size=0,
            product_id="product_id",
            status="active",
        )
        assert_matches_type(SyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    def test_raw_response_list(self, client: DodoPayments) -> None:
        response = client.license_keys.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = response.parse()
        assert_matches_type(SyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    def test_streaming_response_list(self, client: DodoPayments) -> None:
        with client.license_keys.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = response.parse()
            assert_matches_type(SyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

        assert cast(Any, response.is_closed) is True


class TestAsyncLicenseKeys:
    parametrize = pytest.mark.parametrize(
        "async_client", [False, True, {"http_client": "aiohttp"}], indirect=True, ids=["loose", "strict", "aiohttp"]
    )

    @parametrize
    async def test_method_retrieve(self, async_client: AsyncDodoPayments) -> None:
        license_key = await async_client.license_keys.retrieve(
            "lic_123",
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    async def test_raw_response_retrieve(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.license_keys.with_raw_response.retrieve(
            "lic_123",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = await response.parse()
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    async def test_streaming_response_retrieve(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.license_keys.with_streaming_response.retrieve(
            "lic_123",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = await response.parse()
            assert_matches_type(LicenseKey, license_key, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_retrieve(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            await async_client.license_keys.with_raw_response.retrieve(
                "",
            )

    @parametrize
    async def test_method_update(self, async_client: AsyncDodoPayments) -> None:
        license_key = await async_client.license_keys.update(
            id="lic_123",
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    async def test_method_update_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        license_key = await async_client.license_keys.update(
            id="lic_123",
            activations_limit=0,
            disabled=True,
            expires_at=parse_datetime("2019-12-27T18:11:19.117Z"),
        )
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    async def test_raw_response_update(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.license_keys.with_raw_response.update(
            id="lic_123",
        )

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = await response.parse()
        assert_matches_type(LicenseKey, license_key, path=["response"])

    @parametrize
    async def test_streaming_response_update(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.license_keys.with_streaming_response.update(
            id="lic_123",
        ) as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = await response.parse()
            assert_matches_type(LicenseKey, license_key, path=["response"])

        assert cast(Any, response.is_closed) is True

    @parametrize
    async def test_path_params_update(self, async_client: AsyncDodoPayments) -> None:
        with pytest.raises(ValueError, match=r"Expected a non-empty value for `id` but received ''"):
            await async_client.license_keys.with_raw_response.update(
                id="",
            )

    @parametrize
    async def test_method_list(self, async_client: AsyncDodoPayments) -> None:
        license_key = await async_client.license_keys.list()
        assert_matches_type(AsyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    async def test_method_list_with_all_params(self, async_client: AsyncDodoPayments) -> None:
        license_key = await async_client.license_keys.list(
            created_at_gte=parse_datetime("2019-12-27T18:11:19.117Z"),
            created_at_lte=parse_datetime("2019-12-27T18:11:19.117Z"),
            customer_id="customer_id",
            page_number=0,
            page_size=0,
            product_id="product_id",
            status="active",
        )
        assert_matches_type(AsyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    async def test_raw_response_list(self, async_client: AsyncDodoPayments) -> None:
        response = await async_client.license_keys.with_raw_response.list()

        assert response.is_closed is True
        assert response.http_request.headers.get("X-Stainless-Lang") == "python"
        license_key = await response.parse()
        assert_matches_type(AsyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

    @parametrize
    async def test_streaming_response_list(self, async_client: AsyncDodoPayments) -> None:
        async with async_client.license_keys.with_streaming_response.list() as response:
            assert not response.is_closed
            assert response.http_request.headers.get("X-Stainless-Lang") == "python"

            license_key = await response.parse()
            assert_matches_type(AsyncDefaultPageNumberPagination[LicenseKey], license_key, path=["response"])

        assert cast(Any, response.is_closed) is True



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .._models import BaseModel

__all__ = ["AddonCartResponseItem"]


class AddonCartResponseItem(BaseModel):
    """Response struct representing subscription details"""

    addon_id: str

    quantity: int
